import collections, traceback, textwrap

import shapely.geometry, shapely.ops

import esy.osm.pbf


def is_point(entry):
    return type(entry) is esy.osm.pbf.Node


def is_linestring(entry):
    if type(entry) is not esy.osm.pbf.Way: return False
    return entry.tags.get('area', 'no') == 'no'


def is_polygon(entry):
    if type(entry) is not esy.osm.pbf.Way: return False
    if entry.tags.get('area') == 'no': return False
    return entry.refs[0] == entry.refs[-1]


def is_multipolygon(entry):
    if type(entry) is not esy.osm.pbf.Relation: return False
    return entry.tags.get('type') == 'multipolygon'


def merge_touching_rings(rings):
    '''
    Search and merge touching rings.

    OSM polygons deviate from OGC simple features in inner rings of a
    multipolygon. They may touch in OSM but not in OGC.
    '''
    merged = []
    to_merge = list(rings)
    while to_merge:
        a = to_merge.pop(0)
        while True:
            for b in to_merge:
                if a.overlaps(b):
                    a = shapely.ops.unary_union((
                        shapely.geometry.Polygon(a), shapely.geometry.Polygon(b)
                    ))
                    a = shapely.geometry.LinearRing(a.exterior.coords)
                    to_merge.remove(b)
                    break
            else:
                break
        merged.append(a)
    return merged


def merge_segments_to_rings(segments, candidate=None, assigned=None):
    '''
    Iterates all rings that can be merged from segments. Note that rings are not
    necessarily linear.
    '''
    if not candidate:
        candidate, assigned = segments[0], set([0])

    if candidate[0] == candidate[-1]:
        yield candidate, assigned

    for idx, segment in enumerate(segments):
        if idx in assigned:
            continue

        if candidate[-1] == segment[0]:
            yield from merge_segments_to_rings(
                segments, candidate + segment[1:], assigned | set([idx])
            )
        elif candidate[-1] == segment[-1]:
            yield from merge_segments_to_rings(
                segments, candidate + segment[:-1][::-1], assigned | set([idx])
            )


def multipolygon_shape(relation, ways, nodes):
    '''
    Converts an [openstreetmap multipolygon](
    https://wiki.openstreetmap.org/wiki/Relation:multipolygon) to a
    [shapely multipolygon](
    https://shapely.readthedocs.io/en/latest/manual.html#MultiPolygon).
    '''
    geoms = []

    # Collect node references for outer and inner rings.
    rings = dict(outer=[], inner=[])
    for rel_id, role_type, role in relation.members:
        if role_type != 'WAY':
            continue
        if role == '':
            # Assume role is outer.
            role = 'outer'
        rings[role].append(list(ways[rel_id].refs))

    # Merge segments to rings.
    for role, segments in rings.items():
        rings[role] = []
        while segments:
            for candidate, assigned in merge_segments_to_rings(segments):
                ring = shapely.geometry.polygon.LinearRing(
                    [nodes[i].lonlat for i in candidate]
                )
                if ring.is_valid:
                    break
            else:
                raise ValueError('Invalid segments')

            # Prune assigned segments.
            segments = [s for i, s in enumerate(segments) if i not in assigned]

            # Switch winding order.
            if ring.is_ccw != (role == 'outer'):
                ring = shapely.geometry.polygon.LinearRing(ring.coords[::-1])
            assert ring.is_ccw == (role == 'outer')
            rings[role].append(ring)

    for outer in rings['outer']:
        # Construct a polygon from the outer ring to check which inner rings
        # it contains.
        outer = shapely.geometry.Polygon(outer)
        interiors = list(map(outer.contains, rings['inner']))

        holes = [r for r, i in zip(rings['inner'], interiors) if i]

        # Prune inner rings.
        rings['inner'] = [
            r for r, i in zip(rings['inner'], interiors) if not i
        ]

        # Rebuild outer polygon with holes.
        # TODO Is it possible to reuse the original outer polygon?
        shape = shapely.geometry.Polygon(list(outer.exterior.coords), holes)
        if not shape.is_valid:
            # Try to merge touching inner rings (which is an OSM exception).
            holes = merge_touching_rings(holes)
            shape = shapely.geometry.Polygon(list(outer.exterior.coords), holes)
            if not shape.is_valid:
                raise ValueError(
                    'Invalid outer ring in multipolygon relation {}'.format(
                        relation
                    )
                )
        geoms.append(shape)

    return shapely.geometry.MultiPolygon(geoms)


PENDING = object()


class Event(object):
    def __init__(self, context):
        self.context = context
        self.ok = None
        self.value = PENDING
        self.callbacks = []

    def _trigger(self, ok, value):
        if not ok and not self.callbacks and not hasattr(self, 'defused'):
            raise RuntimeError('Unhandled failure') from value
        for callback in self.callbacks:
            self.context.queue.append((callback, ok, value))
        self.ok, self.value, self.callbacks = ok, value, None


class Task(Event):
    def __init__(self, context, generator):
        self.context = context
        self.ok = None
        self.value = PENDING
        self.callbacks = []
        self.generator = generator
        self.context.queue.append((self, True, None))

    def __call__(self, ok, value):
        try:
            while True:
                if ok:
                    event = self.generator.send(value)
                else:
                    event = self.generator.throw(value)

                try:
                    if event.callbacks is not None:
                        event.callbacks.append(self)
                        return
                    ok, value = event, event.value
                except Exception as e:
                    if isinstance(event, Event):
                        raise e
                    ok = False
                    value = RuntimeError('{} is not a Event'.format(event))
        except StopIteration as e:
            self._trigger(True, e.args[0] if e.args else None)
        except RuntimeError as e:
            # RuntimeError are critical errors and their handling cannot be
            # deferred to callbacks.
            raise e
        except Exception as e:
            self._trigger(False, e)


class Context(object):
    current = None

    def __init__(self):
        self.queue = collections.deque()
        self.context = None

    def __enter__(self):
        self.context, Context.current = Context.current, self

    def __exit__(self, exc_type, exc_value, exc_trace):
        self.context, Context.current = None, self.context

    def compute(self, event=None):
        if Context.current is not self:
            raise RuntimeError('Context {} is not active'.format(self))

        if event is None:
            event = Event(self)
            event._trigger(True, None)

        while event.value is PENDING:
            if not self.queue:
                raise RuntimeError('Queue is empty')
            func, *args = self.queue.popleft()
            func(*args)
        return event.value


class Invalid(object):
    def __init__(self, entry, exc_type, exc_args, exc_description):
        self.entry, self.exc_type, self.exc_args, self.exc_description = (
            entry, exc_type, exc_args, exc_description
        )

    def __str__(self):
        return 'Invalid {} (id={})\n{}'.format(
            type(self.entry).__name__, self.entry.id,
            textwrap.indent(self.exc_description, '  ')
        )


def except_invalid(entry, generator):
    try:
        return (yield from generator)
    except Exception as e:
        return (
            Invalid(
                entry, type(e), e.args,
                ''.join(traceback.format_exception(
                    type(e), e, e.__traceback__.tb_next
                ))[:-1]
            ), entry.id, entry.tags
        )


class Shape(object):
    def __init__(self, osmfile, context=None):
        if type(osmfile) is str:
            osmfile = esy.osm.pbf.File(osmfile)
        self.osmfile = osmfile
        if context is None:
            context = Context()
        self.context = context
        self._todo = None
        self.blockmap = {}
        self.blockref = {}
        self._requests = []
        self._requestmap = {}

    def iter_tasks(self, filter=None, max_tasks=2 ** 16):
        tasks = []
        for block in self.osmfile.blocks:
            for entry in block:
                if filter is None or filter(entry):
                    if is_multipolygon(entry):
                        func = self.multipolygon
                    elif is_polygon(entry):
                        func = self.polygon
                    elif is_linestring(entry):
                        func = self.linestring
                    elif is_point(entry):
                        func = self.point
                    else:
                        func = self.unsupported

                    tasks.append(Task(
                        self.context, except_invalid(entry, func(entry))
                    ))

                    if len(tasks) >= max_tasks:
                        yield from tasks
                        tasks = []
        yield from tasks

    def __call__(self, filter=None, max_tasks=2 ** 16):
        with self.context:
            for task in self.iter_tasks(filter, max_tasks):
                yield self.context.compute(task)

    def entries(self, ids):
        event, entries = Event(self.context), []
        if not self._requests:
            self.context.queue.append((self._handle_requests,))
        self._requests.append((event, entries))
        for id in ids:
            events = self._requestmap.get(id)
            if events is None:
                self._requestmap[id] = events = []
            events.append(entries)
        return event

    def _handle_requests(self):
        for entry in self.osmfile:
            requests = self._requestmap.get(entry.id)
            if requests is None:
                continue
            for entries in requests:
                entries.append(entry)

        for event, result in self._requests:
            event._trigger(True, result)
        self._requests, self._requestmap = [], {}

    def unsupported(self, entry):
        description = '{} (id={})'.format(type(entry).__name__, entry.id)
        raise NotImplementedError(description)
        yield

    def point(self, node):
        shape = shapely.geometry.Point(node.lonlat)
        return (shape, node.id, node.tags)
        yield

    def linestring(self, way):
        nodes = {
            entry.id: entry
            for entry in (yield self.entries(way.refs))
            if type(entry) is esy.osm.pbf.Node
        }

        shape = shapely.geometry.LineString([nodes[i].lonlat for i in way.refs])
        return (shape, way.id, way.tags)

    def polygon(self, way):
        nodes = {
            entry.id: entry
            for entry in (yield self.entries(way.refs))
            if type(entry) is esy.osm.pbf.Node
        }

        outer = shapely.geometry.polygon.LinearRing(
            [nodes[i].lonlat for i in way.refs]
        )

        # Switch winding order if necessary.
        if not outer.is_ccw:
            outer = shapely.geometry.polygon.LinearRing(outer.coords[::-1])
        shape = shapely.geometry.Polygon(outer)
        if not shape.is_valid:
            raise ValueError('Invalid polygon {}'.format(way.tags))
        return (shape, way.id, way.tags)

    def multipolygon(self, relation):
        way_ids = set(
            id for id, type, role in relation.members if type == 'WAY'
        )

        ways = {
            entry.id: entry
            for entry in (yield self.entries(way_ids))
            if type(entry) is esy.osm.pbf.Way
        }

        node_ids = set.union(*(set(w.refs) for w in ways.values()))
        nodes = {
            entry.id: entry
            for entry in (yield self.entries(node_ids))
            if type(entry) is esy.osm.pbf.Node
        }

        shape = multipolygon_shape(relation, ways, nodes)
        if not shape.is_valid:
            raise ValueError('Invalid multipolygon relation {}'.format(relation))
        return (shape, relation.id, relation.tags)



import sys, zlib, itertools, pickle
from bisect import bisect

import shapely.geometry

import esy.osm.pbf
from esy.osm.pbf.file import (
    decode_strmap, iter_nodes, iter_ways, iter_relations, Node, Way, Relation
)
from esy.osm.shape.shape import multipolygon_shape


_blob = esy.osm.pbf.fileformat_pb2.Blob()


def parse_header_block(file, block):
    if block.header.type != 'OSMHeader':
        raise ValueError(f'Blob type {block.header.type} is not "OSMHeader"')
    file.seek(block.ofs)
    osmheader = esy.osm.pbf.osmformat_pb2.HeaderBlock()
    osmheader.ParseFromString(esy.osm.pbf.file.read_blob(
        file, block.ofs, block.header.datasize
    ))
    return osmheader


def read_blob2(data):
    '''Read and decompress a openstreetmap blob from `file`.'''
    _blob.ParseFromString(data)
    if _blob.raw:
        return _blob.raw
    elif _blob.zlib_data:
        return zlib.decompress(_blob.zlib_data)
    else:
        raise ValueError('lzma blob not supported')


def unsupported(osm, entry):
    description = '{} (id={})'.format(type(entry).__name__, entry.id)
    return (None, entry.id, {'@error': description})
    #raise NotImplementedError(description)
    yield


def point(osm, node):
    shape = shapely.Point(node.lonlat)
    return (shape, node.id, node.tags)
    yield


def linestring(osm, way):
    shape = shapely.LineString(
        [n.lonlat for n in (yield esy.osm.pbf.Node, way.refs)]
    )
    return (shape, way.id, way.tags)


def polygon(osm, way):
    outer = shapely.LinearRing(
        [n.lonlat for n in (yield esy.osm.pbf.Node, way.refs)]
    )

    # TODO This seems to happen because all ways with the same start and end
    # node are treated as polygons. Falling back to a linestring like this is
    # inefficient.
    if not outer.is_valid:
        print('Invalid polygon {}'.format(way.tags))
        return (shapely.LineString(outer.coords), way.id, way.tags)
        
    # Switch winding order if necessary.
    if not outer.is_ccw:
        outer = shapely.geometry.polygon.LinearRing(outer.coords[::-1])

    shape = shapely.geometry.Polygon(outer)
    if not shape.is_valid:
        print(list(shape.exterior.coords))
        print('Invalid polygon {}'.format(way.tags))
        return (None, way.id, way.tags)
        #raise ValueError('Invalid polygon {}'.format(way.tags))
    return (shape, way.id, way.tags)


def multipolygon(osm, relation):
    way_ids = set(
        id for id, type, role in relation.members if type == 'WAY'
    )

    ways = {way.id: way for way in (yield esy.osm.pbf.Way, way_ids)}

    node_ids = set.union(*(set(w.refs) for w in ways.values()))
    nodes = {entry.id: entry for entry in (yield esy.osm.pbf.Node, node_ids)}

    shape = multipolygon_shape(relation, ways, nodes)
    if not shape.is_valid:
        raise ValueError('Invalid multipolygon relation {}'.format(relation))
    return (shape, relation.id, relation.tags)


def shape_function(entry):
    entry_type = type(entry)
    if entry_type is esy.osm.pbf.Node:
        return point
    elif entry_type is esy.osm.pbf.Way:
        if entry.tags.get('area') != 'no' and entry.refs[0] == entry.refs[-1]:
            return polygon
        else:
            return linestring
    elif entry_type is esy.osm.pbf.Relation:
        if entry.tags.get('type') == 'multipolygon':
            return multipolygon
    return unsupported


#entry_typeids = {
#    esy.osm.pbf.Node: 0, esy.osm.pbf.Way: 1, esy.osm.pbf.Relation: 2
#}

osmtypeidxmap = {
    type: idx for idx, type in enumerate(
        (esy.osm.pbf.Node, esy.osm.pbf.Way, esy.osm.pbf.Relation)
    )
}


class Shape(object):
    # TODO Maybe generalize task batching and expose as general scheme to access
    # OSM data? Extracting non-shape relations might be handy.

    def __init__(self, osm):
        if type(osm) is str:
            osm = esy.osm.pbf.Index(osm)
        self.osm = osm

        header = parse_header_block(self.osm.pbf.file, next(self.osm.pbf.blocks))
        self.box = shapely.geometry.box(
            header.bbox.left / 1000000000, header.bbox.bottom / 1000000000,
            header.bbox.right / 1000000000, header.bbox.top / 1000000000
        )

    def __call__(self, filter=None, max_tasks=2 ** 16):
        tasks = []
        for entry in self.osm:
            if filter is not None and not filter(entry):
                continue

            tasks.append(shape_function(entry)(self, entry))
            if len(tasks) < max_tasks:
                continue

            yield from self._process(tasks)
            del tasks[:]

        yield from self._process(tasks)

    def load(self, entries):
        return self._process([
            shape_function(entry)(self, entry) for entry in entries
        ])

    def _process(self, tasks):
        results, queries, requests = {}, (set(), set(), set()), []
        queue = [(generator, True, None) for generator in tasks]
        while queue:
            for generator, ok, value in queue:
                try:
                    if ok:
                        osmtype, ids = generator.send(value)
                    else:
                        osmtype, ids = generator.throw(value)
                except StopIteration as e:
                    results[generator] = e.args[0]
                    continue

                osmtypeidx = osmtypeidxmap[osmtype]
                queries[osmtypeidx].update(ids)
                requests.append((generator, osmtypeidx, ids))

            del queue[:]
            if not requests:
                continue

            # Query data.
            dataset = self.osm(queries)
            for generator, osmtypeidx, ids in requests:
                try:
                    map = dataset[osmtypeidx]
                    queue.append((generator, True, [map[id] for id in ids]))
                except KeyError as error:
                    import traceback
                    traceback.print_exc()
                    queue.append((generator, False, error))

            # Release memory.
            del dataset
            del requests[:]
            for query in queries:
                query.clear()

        for task in tasks:
            yield results[task]



import shapely
import matplotlib.collections
from matplotlib.path import Path


def empty_style():
    empty = {}
    while True:
        yield empty


def render(shapes, path_patch_options=None, point_radius=0.0001):
    if path_patch_options is None:
        path_patch_options = empty_style()
    pathpatches = []
    for shape, path_patch_option in zip(shapes, path_patch_options):
        if type(shape) is shapely.geometry.LineString:
            pathpatches.append(matplotlib.patches.PathPatch(
                matplotlib.path.Path(list(shape.coords), closed=False),
                **{'fill': False, **path_patch_option},
            ))
        elif type(shape) is shapely.geometry.LinearRing:
            pathpatches.append(matplotlib.patches.PathPatch(
                matplotlib.path.Path(list(shape.coords), closed=True),
                **{'fill': False, **path_patch_option},
            ))
        elif type(shape) is shapely.geometry.Polygon:
            vertices = list(shape.exterior.coords)
            codes = (
                [Path.MOVETO] +
                [Path.LINETO] * (len(vertices) - 2) +
                [Path.CLOSEPOLY]
            )

            for ring in shape.interiors:
                ring_vertices = list(ring.coords)
                ring_codes = (
                    [Path.MOVETO] +
                    [Path.LINETO] * (len(ring_vertices) - 2) +
                    [Path.CLOSEPOLY]
                )
                vertices += ring_vertices
                codes += ring_codes
            pathpatches.append(matplotlib.patches.PathPatch(
                matplotlib.path.Path(vertices, codes, closed=True),
                **{'fill': True, **path_patch_option},
            ))
        elif type(shape) is shapely.geometry.MultiPolygon:
            for shape in shape.geoms:
                vertices = list(shape.exterior.coords)
                codes = (
                    [Path.MOVETO] +
                    [Path.LINETO] * (len(vertices) - 2) +
                    [Path.CLOSEPOLY]
                )

                for ring in shape.interiors:
                    ring_vertices = list(ring.coords)
                    ring_codes = (
                        [Path.MOVETO] +
                        [Path.LINETO] * (len(ring_vertices) - 2) +
                        [Path.CLOSEPOLY]
                    )
                    vertices += ring_vertices
                    codes += ring_codes
                pathpatches.append(matplotlib.patches.PathPatch(
                    matplotlib.path.Path(vertices, codes, closed=True),
                    **{'fill': True, **path_patch_option},
                ))
        elif type(shape) is shapely.geometry.Point:
            pathpatches.append(matplotlib.patches.PathPatch(
                matplotlib.path.Path.circle(
                    center=(shape.x, shape.y), radius=point_radius
                ),
                **{'edgecolor': 'none', **path_patch_option},
            ))
        else:
            raise ValueError('Unsupported shape type {}'.format(shape))

    return pathpatches


def patches(shapes, path_patch_options=None):
    return matplotlib.collections.PatchCollection(
        render(shapes, path_patch_options), match_original=True
    )



import os, re, setuptools


def parse_version(relpath):
    path = os.path.join(os.path.abspath(os.path.dirname(__file__)), relpath)
    with open(path, 'r') as f:
        return re.search(
            '^__version__ = \'(?P<version>.*?)\'$', f.read(), re.MULTILINE
        ).group('version')


setuptools.setup(
    name='esy-osm-shape',
    description='Convert OpenStreetMap objects to shapely objects.',
    version=parse_version('src/esy/osm/shape/__init__.py'),
    license='GPLv3',
    long_description=open('README.md', 'r').read(),
    long_description_content_type='text/markdown',
    author='Ontje Lünsdorf',
    author_email='ontje.luensdorf@dlr.de',
    package_dir={'': 'src'},
    packages=setuptools.find_namespace_packages(where='src'),
    python_requires='>= 3.5',
    install_requires=['shapely >= 1.6', 'esy-osm-pbf == 0.0.1'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        (
            'License :: OSI Approved :: '
            'GNU General Public License v3 or later (GPLv3+)'
        ),
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering',
    ],
)
